//
//  basic example:
//  - basic types and enum
//  - std::string
//  - list of plain objects
//  jsonserial - Eric Lecolinet
// See: https://www.telecom-paris.fr/~elc/jsonserial/
//

#include "main.hpp"
#include "../jsonserial.hpp"
#include "../jsonserial_list.hpp"
using namespace jsonserial;

namespace basics {

class Contact {
public:
  enum Gender {Unknown, Male, Female, Other};
  struct Phone {std::string type, number;};

  friend JSonClasses* registerClasses();  // needed if variables are not public!

  Contact() = default;                        // needed to instantiate pointees!

  Contact(std::string const& firstname, std::string const& lastname,
          unsigned int age, Gender gender)
  : firstname(firstname), lastname(lastname), age(age), gender(gender) {}

  void addPhone(std::string const& type, std::string const& number) {
    phones.push_back(Phone{type, number});
  }
  
  void print(std::ostream& out) const {
    out
    << "First Name:\t" << firstname << "\n"
    << "Last Name:\t"  << lastname  << "\n"
    << "Gender:\t"     << gender    << "\n"
    << "Age:\t"        << age << "\n"
    << "Phones:\t";
    for (auto& it : phones) {
      out << "[" << it.type << ": " << it.number << "] ";
    }
    out << "\n";
  }

private:
  std::string firstname, lastname;
  unsigned int age{};
  Gender gender;
  std::list<Phone> phones;  // list of plain objects
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

using Contacts = std::list<Contact*>;

void printList(std::string const& title, Contacts const& list) {
  std::cout << "--- " << title << " --------\n\n";
  for (auto& it : list) {it->print(std::cout); std::cout << std::endl;}
}

Contacts createList() {
  Contact* bob = new Contact("Bob", "Dûpontèle", 31, Contact::Male);
  bob->addPhone("home", "123 456-7890");
  bob->addPhone("mobile","703 221-2121");

  Contact* alice = new Contact("Alice", "Dûpontèle", 33, Contact::Female);
  alice->addPhone("home", "123 456-7890");
  alice->addPhone("office", "211 1234-1234");

  return Contacts{bob, alice};
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

JSonClasses* registerClasses() {
  auto* classes = new JSonClasses();

  classes->addClass<Contact>("Contact")
  .addMember("firstname", &Contact::firstname)
  .addMember("lastname", &Contact::lastname)
  .addMember("gender", &Contact::gender)
  .addMember("age", &Contact::age)
  .addMember("phones", &Contact::phones);

  classes->addClass<Contact::Phone>("Phone")
  .addMember("type", &Contact::Phone::type)
  .addMember("number", &Contact::Phone::number);

  return classes;
}


static JSonSerial js(registerClasses());


void test_file(std::string const& filename) {
  Contacts list = createList();
  printList("Original List", list);

  // last argument can be false because the objects' graph is not cyclic!
  if (!js.write(list, filename, false)) return;

  Contacts list2;
  if (!js.read(list2, filename)) return;
  printList("Copy after Reading", list2);
}


void test_stream() {
  Contacts list = createList();
  printList("Original List", list);

  std::stringstream buffer;

  // last argument can be false because the objects' graph is not cyclic!
  if (!js.write(list, buffer, false)) return;
  std::cout << buffer.str() << std::endl;

  Contacts list2;
  if (!js.read(list2, buffer)) return;
  printList("Copy after Reading", list2);
}

}
