//
//  Custom class/member registration:
//  - missing no-argument constructor
//  - registration of members through their accessors
//  - custom functions for reading/writing members (missing appropriate set function)
//  jsonserial - Eric Lecolinet
// See: https://www.telecom-paris.fr/~elc/jsonserial/
//

#include "main.hpp"
#include "../jsonserial.hpp"
#include "../jsonserial_list.hpp"
#include "../jsonserial_map.hpp"
using namespace jsonserial;

namespace misc_registration {

class Contact {
public:
  enum Gender {Unknown, Male, Female, Other};
  struct Phone {std::string type, number;};

  // NOTES:
  // - registerClasses() is not declared as a friend, accessors will be used.
  // - the class does not have a no-argument constructor

  Contact(std::string const& firstname, std::string const& lastname,
          unsigned int age, Gender gender)
  : firstname(firstname), lastname(lastname), age(age), gender(gender) {}

  void setFirstName(std::string const& name) {firstname = name;}
  std::string const& getFirstName() const {return firstname;}

  void setLastName(std::string const& name) {lastname = name;}
  std::string const& getLastName() const {return lastname;}

  void setGender(Gender value) {gender = value;}
  Gender getGender() const {return gender;}

  void setAge(unsigned int value) {age = value;}
  unsigned int const& getAge() const {return age;}

  void setPartner(Contact* p) {partner = p;}
  Contact* getPartner() const {return partner;}

  void setParent1(Contact* p) {parent1 = p;}
  Contact* getParent1() const {return parent1;}

  void setParent2(Contact* p) {parent2 = p;}
  Contact* getParent2() const {return parent2;}

  void addPhone(std::string const& type, std::string const& number) {
    phones.push_back(Phone{type, number});
  }

  void setPhones(std::list<Phone> const& phones) {
    this->phones = phones;
  }

  std::list<Phone> const& getPhones() const {
    return phones;
  }

  void addChild(Contact* child) {
    if (child) children[child->firstname] = child;
  }

  // NOTE: no setChildren() method!

  std::map<std::string,Contact*> const& getChildren() const {
    return children;
  }

  void print(std::ostream& out) const {
    out
    << "First:\t" << firstname << "\n"
    << "Last:\t"  << lastname  << "\n"
    << "Gender:\t"<< gender << "\n"
    << "Age:\t" << age << "\n"
    << "Phones:\t";
    for (auto& it : phones) {
      out << "[" << it.type << ": " << it.number << "] ";
    }
    out << "\n";
    if (partner) out << "Partner:\t"<< partner->firstname << "\n";
    if (parent1) out << "Parent1:\t"<< parent1->firstname << "\n";
    if (parent2) out << "Parent2:\t"<< parent2->firstname << "\n";
    if (!children.empty()) {
      out << "Children:\t";
      for (auto& it : children) out << it.second->firstname << " ";
      out << "\n";
    }
  }

private:
  std::string firstname, lastname;
  Gender gender;
  unsigned int age{};
  std::list<Phone> phones;
  std::map<std::string, Contact*> children; 
  Contact *partner{}, *parent1{}, *parent2{}; // NOTE: pointers must be initialized
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

using Contacts = std::list<Contact*>;

void printList(std::string const& title, Contacts const& list) {
  std::cout << "--- " << title << " --------\n\n";
  for (auto& it : list) {it->print(std::cout); std::cout << std::endl;}
}

Contacts createList() {
  Contact* bob = new Contact("Bob", "Dûpontèle", 31, Contact::Male);
  bob->addPhone("home", "123 456-7890");
  bob->addPhone("mobile","703 221-2121");

  Contact* alice = new Contact("Alice", "Dûpontèle", 33, Contact::Female);
  alice->addPhone("home", "123 456-7890");
  alice->addPhone("office", "211 1234-1234");

  Contact* karim = new Contact("Karim", "Dûpontèle", 9, Contact::Male);
  karim->addPhone("mobile","122 122-1222");

  Contact* susan = new Contact("Susan", "Dûpontèle", 11, Contact::Female);
  susan->addPhone("mobile","133 133-1333");

  bob->setPartner(alice);
  bob->addChild(karim);
  bob->addChild(susan);

  alice->setPartner(bob);
  alice->addChild(karim);
  alice->addChild(susan);

  karim->setParent1(bob);
  karim->setParent2(alice);

  susan->setParent1(bob);
  susan->setParent2(alice);

  return Contacts{bob, alice, karim, susan};
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

JSonClasses* registerClasses() {
  JSonClasses* classes = new JSonClasses();

  // NOTE: a creator function is provided because the class has no no-argument constructor
  classes->addClass<Contact>("Contact", [](){
    return new Contact("", "", 0, Contact::Unknown);
  })
  .addMember("firstname", &Contact::setFirstName, &Contact::getFirstName)
  .addMember("lastname", &Contact::setLastName, &Contact::getLastName)
  .addMember("gender", &Contact::setGender, &Contact::getGender)
  .addMember("age", &Contact::setAge,  &Contact::getAge)
  .addMember("phones", &Contact::setPhones, &Contact::getPhones)
  .addMember("partner", &Contact::setPartner, &Contact::getPartner)
  .addMember("parent1", &Contact::setParent1, &Contact::getParent1)
  .addMember("parent2", &Contact::setParent2, &Contact::getParent2)
  // NOTE: the class does not have a setChildren() method, thus we must write
  // a custom function to read and write children
  .addMember("children",
       [](Contact& c, JSonSerial& js, std::string const& value){  // read children
    std::map<std::string, Contact*> children;
    js.readMember(children, value);
    for (auto& it : children) c.addChild(it.second);
  },
       [](const Contact& c, JSonSerial& js){ // write children
    js.writeMember(c.getChildren());
  });

  classes->addClass<Contact::Phone>("Phone")
  .addMember("type", &Contact::Phone::type)
  .addMember("number", &Contact::Phone::number);

  return classes;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static JSonSerial js(registerClasses());


void test_file(std::string const& filename) {
  Contacts list = createList();
  printList("Original List", list);

  // CAUTION: last argument must be true because the objects' graph is cyclic!
  if (!js.write(list, filename, true)) return;

  Contacts list2;
  if (!js.read(list2, filename)) return;
  printList("Copy after Reading", list2);
}


void test_stream() {
  Contacts list = createList();
  printList("Original List", list);

  std::stringstream buffer;

  // CAUTION: last argument must be true because the objects' graph is cyclic!
  if (!js.write(list, buffer, true)) return;
  std::cout << buffer.str() << std::endl;

  Contacts list2;
  if (!js.read(list2, buffer)) return;
  printList("Copy after Reading", list2);
}

}
